"""Test data helpers."""

from unittest.mock import Mock

from pyfibaro.fibaro_data_helper import (
    read_devices,
    read_rooms,
    find_master_devices,
)
from pyfibaro.fibaro_device import DeviceModel
from pyfibaro.fibaro_room import RoomModel

from .test_utils import load_fixture

scene_payload = load_fixture("scene-hc3.json")
room_payload = load_fixture("room.json")
device_payload = load_fixture("device-hc3.json")
device2_payload = load_fixture("device.json")
device3_payload = load_fixture("device-netatmo-plugin.json")


def test_read_rooms() -> None:
    """Test read rooms"""
    client = Mock()
    client.read_rooms.return_value = [RoomModel(room_payload[0])]

    rooms = read_rooms(client)

    assert len(rooms) == 1
    assert rooms[4] == "Wohnen"


def test_read_devices_main_devices() -> None:
    """Test read devices"""
    devices = [
        # z-wave controller is no device
        DeviceModel(device_payload[0], Mock(), 4),
        # main endpoint reported as device
        DeviceModel(device2_payload[2], Mock(), 4),
        # child endpoint not reported as device
        DeviceModel(device2_payload[3], Mock(), 4),
        # zigbee controller is no device
        DeviceModel(device_payload[4], Mock(), 4),
    ]
    client = Mock()
    client.read_devices.return_value = devices

    devices = find_master_devices(devices)

    assert len(devices) == 1
    assert devices[0].name == "Hifi Main"


def test_read_devices_main_devices_without_zwave_controller() -> None:
    """Test read z-wave devices without a zwave controller"""
    devices = [
        # main entity
        DeviceModel(device2_payload[2], Mock(), 4),
        # child entity
        DeviceModel(device2_payload[3], Mock(), 4),
    ]
    client = Mock()
    client.read_devices.return_value = devices

    devices = find_master_devices(devices)

    assert len(devices) == 1
    assert devices[0].name == "Hifi Main"


def test_read_devices_main_devices_netatmo() -> None:
    """Test read netatmo devices"""
    devices = [
        DeviceModel(device, Mock(), 4) for device in device3_payload
    ]
    client = Mock()
    client.read_devices.return_value = devices

    devices = find_master_devices(devices)

    assert len(devices) == 2
    assert devices[0].name == "Netatmo Weather Sation"
    assert devices[1].name == "Netatmo Weather Station 2"


def test_read_devices() -> None:
    """Test read devices"""
    devices = [
        DeviceModel(device2_payload[2], Mock(), 4),
        DeviceModel(device2_payload[3], Mock(), 4),
    ]
    client = Mock()
    client.read_devices.return_value = devices

    devices = read_devices(client, True)

    assert len(devices) == 2



"""Fibaro data helper provides static method to read and
process data for the different fibaro API endpoints."""

from pyfibaro.fibaro_client import FibaroClient
from pyfibaro.fibaro_device import DeviceModel

ZWAVE_CONTROLLER = "com.fibaro.zwavePrimaryController"

CONTROLLER_TYPES = [
    ZWAVE_CONTROLLER,
    "com.fibaro.zigbeePrimaryController",
    "com.fibaro.niceEngine",
]


def read_rooms(fibaro_client: FibaroClient) -> dict[int, str]:
    """Read dictionary mapping room ids to room name."""
    return {room.fibaro_id: room.name for room in fibaro_client.read_rooms()}


def read_devices(
    fibaro_client: FibaroClient, include_devices_from_plugins: bool = False
) -> list[DeviceModel]:
    """Read all enabled devices."""
    devices = fibaro_client.read_devices()

    return [
        device
        for device in devices
        if (not device.is_plugin or include_devices_from_plugins)
        and device.enabled
    ]


def find_master_devices(devices: list[DeviceModel]) -> list[DeviceModel]:
    """Find main devices only."""
    controller_ids = _get_controller_ids(devices)

    return _main_devices(devices, controller_ids)


def _get_controller_ids(devices: list[DeviceModel]) -> set[int]:
    controller_ids = {
        device.fibaro_id for device in devices if device.type in CONTROLLER_TYPES}
    zwave_controller_id = {
        device.fibaro_id for device in devices if device.type == ZWAVE_CONTROLLER}
    if len(zwave_controller_id) == 0:
        # When choosing a user account without admin rights sometimes the controller device
        # is not visible on the API. In this case we assume it has id = 1.
        controller_ids.add(1)
    return controller_ids


def _main_devices(devices: list[DeviceModel], controller_ids: set[int]) -> bool:
    children_by_parent_id: dict[int, list[DeviceModel]] = {}
    for d in devices:
        children = children_by_parent_id.setdefault(d.parent_fibaro_id, [])
        children.append(d)

    main_devices = []

    for d in devices:
        if d.parent_fibaro_id in controller_ids or (
            d.parent_fibaro_id == 0 and d.fibaro_id not in controller_ids
        ):
            # Normally devices hanging on a controller or on nothing (id=0)
            # are representing a device.
            # Speical case are plugins which sometimes have an additional parent
            # representing the plugin itself, so devices are the first child level.
            count_of_children_level = 0
            direct_children = children_by_parent_id.get(d.fibaro_id)
            if direct_children:
                count_of_children_level = 1
                for child in direct_children:
                    if children_by_parent_id.get(child.fibaro_id):
                        count_of_children_level = 2

            if count_of_children_level == 2:
                main_devices.extend(direct_children)
            else:
                main_devices.append(d)
    return main_devices



"""Test FibaroInfo class."""

import requests_mock

from pyfibaro.fibaro_client import FibaroClient

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture


def test_fibaro_connect_success() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        login_payload = load_fixture("login_success.json")
        info_payload = load_fixture("info.json")

        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)
        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        logged_in = client.connect()

        assert logged_in is True
        assert mock.call_count == 2


def test_fibaro_connect_fail() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        login_payload = load_fixture("login_fail.json")
        info_payload = load_fixture("info.json")

        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)
        client = FibaroClient(TEST_BASE_URL)
        logged_in = client.connect()

        assert logged_in is False
        assert mock.call_count == 2



"""Helpers for fibaro tests."""

import json
from typing import Any

TEST_BASE_URL = "http://192.168.1.166/api/"
TEST_USERNAME = "admin"
TEST_PASSWORD = "admin"


def load_fixture(filename: str) -> Any:
    """Load a json file."""
    with open(f"tests/fixture/{filename}", encoding="UTF-8") as file:
        return json.load(file)



"""Test FibaroDeviceManager."""

from unittest.mock import Mock

from pyfibaro.fibaro_device import DeviceModel
from pyfibaro.fibaro_device_manager import FibaroDeviceManager

from .test_utils import load_fixture

device_payload = load_fixture("device.json")


def test_fibaro_device_manager_get_devices() -> None:
    """Test fibaro device manager."""
    devices = [
        DeviceModel(device_payload[2], Mock(), 4),
        DeviceModel(device_payload[3], Mock(), 4),
    ]

    fibaro_client = Mock()
    fibaro_client.read_devices.return_value = devices

    multiplexer = FibaroDeviceManager(fibaro_client)
    devices = multiplexer.get_devices()
    assert len(devices) == 2


def test_fibaro_device_manager_add_state_listener() -> None:
    """Test manager add state change listener."""
    devices = [
        DeviceModel(device_payload[2], Mock(), 4),
        DeviceModel(device_payload[3], Mock(), 4),
    ]

    listener = Mock()

    fibaro_client = Mock()
    fibaro_client.read_devices.return_value = devices

    manager = FibaroDeviceManager(fibaro_client)
    remove = manager.add_change_listener(13, listener.call_method)

    remove()

    manager.close()


def test_fibaro_device_manager_add_event_listener() -> None:
    """Test manager add event listener."""
    devices = [
        DeviceModel(device_payload[2], Mock(), 4),
        DeviceModel(device_payload[3], Mock(), 4),
    ]

    listener = Mock()

    fibaro_client = Mock()
    fibaro_client.read_devices.return_value = devices

    manager = FibaroDeviceManager(fibaro_client)
    remove = manager.add_event_listener(28, listener.call_method)

    remove()



"""Test FibaroInfo class."""

import requests_mock

from pyfibaro.fibaro_client import FibaroClient

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture

device_payload = load_fixture("device-scene-support.json")
login_payload = load_fixture("login_success.json")
info_payload = load_fixture("info.json")


def test_fibaro_device_scene_support() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()
        assert devices is not None

        assert len(devices) == 3

        assert devices[2].has_central_scene_event is True
        # Invalid value returns False
        central_scene_event = devices[2].central_scene_event
        assert len(central_scene_event) == 4
        assert central_scene_event[0].key_id == 1
        assert central_scene_event[0].key_event_types == [
            "Pressed",
            "Released",
            "HeldDown",
        ]

        assert devices[1].has_central_scene_event is False
        assert devices[1].central_scene_event == []



"""Endpoint object to access the endpoint settings/info"""
from __future__ import annotations

import json
import logging
from typing import Any

from .common.const import IGNORE_DEVICE
from .common.rest_client import RestClient

_LOGGER = logging.getLogger(__name__)


def _to_bool(bool_value: bool | str) -> bool:
    """Convert any value to bool."""
    if bool_value is not None:
        if isinstance(bool_value, bool):
            return bool_value
        if isinstance(bool_value, str):
            return bool_value.lower() == "true"
    return False


class DeviceModel:
    """Model of a device."""

    def __init__(self, data: dict, rest_client: RestClient, api_version: int) -> None:
        """Constructor."""
        self.raw_data = data
        self._rest_client = rest_client
        self._api_version = api_version

    @property
    def fibaro_id(self) -> int:
        """Device id"""
        return self.raw_data.get("id")

    @property
    def name(self) -> str:
        """Device name"""
        return self.raw_data.get("name")

    @property
    def parent_fibaro_id(self) -> int:
        """Id of the parent device or 0 if there is no parent."""
        return int(self.raw_data.get("parentId", 0))

    @property
    def type(self) -> str | None:
        """Device type."""
        return self.raw_data.get("type")

    @property
    def base_type(self) -> str | None:
        """Device base type."""
        return self.raw_data.get("baseType")

    @property
    def room_id(self) -> int:
        """Room id of the device or 0 if no room is assigned."""
        return int(self.raw_data.get("roomID", 0))

    @property
    def properties(self) -> dict:
        """Get the properties."""
        return self.raw_data.get("properties", {})

    @property
    def actions(self) -> dict[str, int]:
        """Get the available actions."""
        return self.raw_data.get("actions", {})

    def has_interface(self, interface_name: str) -> bool:
        """Returns True if the device has the according interface defined.

        Interfaces are capabilities of the device.
        """
        return interface_name in self.raw_data.get("interfaces", [])

    @property
    def unit(self) -> str | None:
        """Returns the unit of the devices value attribute or None."""
        return self.properties.get("unit")

    @property
    def has_unit(self) -> bool:
        """Returns true if the device has a unit property."""
        return "unit" in self.properties

    @property
    def endpoint_id(self) -> int:
        """Returns the endpoint id or 0 if there is no endpoint id.

        Endpoints are numbered if one physical device exposes different endpoints.
        """
        return int(self.properties.get("endPointId", 0))

    @property
    def has_endpoint_id(self) -> bool:
        """Returns true if the device has a unit property."""
        return "endPointId" in self.properties

    @property
    def enabled(self) -> bool:
        """Returns the enabled flag of the device.

        If the device does not support that flag, True is returned.
        """
        return self.raw_data.get("enabled", True)

    @property
    def visible(self) -> bool:
        """Returns the visible state of the device."""
        return self.raw_data.get("visible", True)

    @property
    def is_plugin(self) -> bool:
        """Returns the device type.

        Returns:
        True for virtual devices and Quick Apps
        False for physical devices and controllers
        """
        return self.raw_data.get("isPlugin", True)

    @property
    def battery_level(self) -> int:
        """Returns the battery level of the device in percent."""
        percent = int(self.properties.get("batteryLevel", 0))
        return 0 if percent == 255 else percent

    @property
    def has_battery_level(self) -> bool:
        """Returns true if the device has a unit property."""
        return "batteryLevel" in self.properties

    @property
    def armed(self) -> bool:
        """Returns the armed state of the device if supported,
        otherwise False is returned.
        """
        return _to_bool(self.properties.get("armed", False))

    @property
    def has_armed(self) -> bool:
        """Returns true if the device has a unit property."""
        return "armed" in self.properties

    @property
    def dead(self) -> bool:
        """Returns the state if the device is reachable if supported,
        otherwise False is returned.
        """
        return _to_bool(self.properties.get("dead", False))

    @property
    def has_dead(self) -> bool:
        """Returns true if the device has a dead property."""
        return "dead" in self.properties

    @property
    def dead_reason(self) -> str | None:
        """Returns the dead reason or None if not supported."""
        return self.properties.get("deadReason")

    @property
    def has_dead_reason(self) -> bool:
        """Returns true if the device has a deadReason property."""
        return "deadReason" in self.properties

    @property
    def value(self) -> ValueModel:
        """Returns the value info."""
        return ValueModel(self.properties, "value")

    @property
    def value_2(self) -> ValueModel:
        """Returns the value info."""
        return ValueModel(self.properties, "value2")

    @property
    def state(self) -> ValueModel:
        """Returns the state info."""
        return ValueModel(self.properties, "state")

    @property
    def color(self) -> ColorModel:
        """Returns the color info."""
        return ColorModel(self.properties, "color")

    @property
    def last_color_set(self) -> ColorModel:
        """Returns the last set color info."""
        return ColorModel(self.properties, "lastColorSet")

    @property
    def brightness(self) -> int:
        """Returns the brightness of the device. If missing, 0 is returned."""
        return int(self.properties.get("brightness", 0))

    @property
    def has_brightness(self) -> bool:
        """Returns true if the device has a brightness property."""
        return "brightness" in self.properties

    @property
    def current_program(self) -> int:
        """Returns the current program of the device or None."""
        return int(self.properties.get("currentProgram", 0))

    @property
    def current_program_id(self) -> int:
        """Returns the current pogram id of the device or 0."""
        return int(self.properties.get("currentProgramID", 0))

    @property
    def mode(self) -> int:
        """Returns the mode or 0 if there is no mode."""
        return int(self.properties.get("mode", 0))

    @property
    def has_mode(self) -> bool:
        """Returns true if the device has a mode."""
        return "mode" in self.properties

    @property
    def supported_modes(self) -> list[int]:
        """Returns the supported modes, for example for fan or hvac devices."""
        if "supportedModes" in self.properties:
            modes = self.properties.get("supportedModes")
            if isinstance(modes, str) and modes != "":
                return [int(mode) for mode in modes.split(",")]
            if isinstance(modes, list):
                return [int(mode) for mode in modes]
        return []

    @property
    def has_supported_modes(self) -> bool:
        """Returns true if the device has a supported modes property."""
        return "supportedModes" in self.properties

    @property
    def operating_mode(self) -> int:
        """Returns the operating mode or 0 if there is no operating mode."""
        return int(self.properties.get("operatingMode", 0))

    @property
    def has_operating_mode(self) -> bool:
        """Returns true if the device has a operating mode."""
        return "operatingMode" in self.properties

    @property
    def supported_operating_modes(self) -> list[int]:
        """Returns the supported operating modes, for example for fan or hvac devices."""
        if "supportedOperatingModes" in self.properties:
            modes = self.properties.get("supportedOperatingModes")
            if isinstance(modes, str) and modes != "":
                return [int(mode) for mode in modes.split(",")]
            if isinstance(modes, list):
                return [int(mode) for mode in modes]
        return []

    @property
    def has_supported_operating_modes(self) -> bool:
        """Returns true if the device has a supported operating modes property."""
        return "supportedOperatingModes" in self.properties

    @property
    def thermostat_mode(self) -> str | None:
        """Returns the thermostat mode or None if there is no thermostat mode."""
        return self.properties.get("thermostatMode")

    @property
    def has_thermostat_mode(self) -> bool:
        """Returns true if the device has a thermostat mode."""
        return "thermostatMode" in self.properties

    @property
    def thermostat_operating_state(self) -> str | None:
        """Returns the thermostat operating state or None if there
        is no thermostat operating state.
        """
        return self.properties.get("thermostatOperatingState")

    @property
    def has_thermostat_operating_state(self) -> bool:
        """Returns true if the device has a thermostat operating state."""
        return "thermostatOperatingState" in self.properties

    @property
    def supported_thermostat_modes(self) -> list[str]:
        """Returns the supported thermostat modes, for example for hvac devices."""
        if "supportedThermostatModes" in self.properties:
            return self.properties.get("supportedThermostatModes")
        return []

    @property
    def has_supported_thermostat_modes(self) -> bool:
        """Returns true if the device has a supported thermostat modes property."""
        return "supportedThermostatModes" in self.properties

    @property
    def heating_thermostat_setpoint(self) -> float:
        """Returns the heating thermostat setpoint or 0 if there is no value."""
        return float(self.properties.get("heatingThermostatSetpoint", 0.0))

    @property
    def has_heating_thermostat_setpoint(self) -> bool:
        """Returns true if the device has a heating thermostat setpoint property."""
        return "heatingThermostatSetpoint" in self.properties

    @property
    def heating_thermostat_setpoint_future(self) -> float:
        """Returns the heating thermostat setpoint future or 0 if there is no value."""
        return float(self.properties.get("heatingThermostatSetpointFuture", 0.0))

    @property
    def has_heating_thermostat_setpoint_future(self) -> bool:
        """Returns true if the device has a heating thermostat setpoint future property."""
        return "heatingThermostatSetpointFuture" in self.properties

    @property
    def target_level(self) -> float:
        """Returns the target level or 0 if there is no value."""
        return float(self.properties.get("targetLevel", 0.0))

    @property
    def has_central_scene_event(self) -> bool:
        """Returns true if the device can issue central scene events."""
        return "centralSceneSupport" in self.properties

    @property
    def central_scene_event(self) -> list[SceneEvent]:
        """Returns list of potential scene events."""
        central_scene_support = []
        value = self.properties.get("centralSceneSupport")
        if isinstance(value, list):
            central_scene_support = value
        if isinstance(value, str):
            central_scene_support = json.loads(value)

        result = []
        for central_scene in central_scene_support:
            key_id = int(central_scene.get("keyId"))
            key_attributes = central_scene.get("keyAttributes")
            result.append(SceneEvent(key_id, key_attributes))
        return result

    def execute_action(self, action: str, arguments: list[Any] | None = None) -> Any:
        """Execute a device action.

        Params:
        action: name of the action to call
        arguments: list of arguments needed for the action
        """
        if action not in self.actions:
            _LOGGER.warning(
                "The device %s has no action %s. Possible actions are %s",
                self.fibaro_id,
                action,
                self.actions,
            )

        url = f"devices/{self.fibaro_id}/action/{action}"

        args_prepared = {"args": arguments} if arguments else {}
        _LOGGER.debug(
            "Execute %s for device %s with args %s.",
            action,
            self.fibaro_id,
            args_prepared,
        )
        return self._rest_client.post(url, json=args_prepared)

    @staticmethod
    def read_devices(rest_client: RestClient, api_version: int) -> list[DeviceModel]:
        """Returns a list of devices."""
        raw_data: list[dict] = rest_client.get("devices")

        devices: list[dict] = []
        for device in raw_data:
            if device.get("type") in IGNORE_DEVICE:
                _LOGGER.debug("Ignore device: %s", device.get("id"))
            elif "id" not in device or "name" not in device:
                _LOGGER.debug(
                    "Ignore device because it does not contain id or name")
            else:
                devices.append(device)
        return [DeviceModel(data, rest_client, api_version) for data in devices]


class ValueModel:
    """Model to read out the value in several ways."""

    def __init__(self, properties: dict, property_name: str) -> None:
        """Constructor."""
        self._properties = properties
        self._property_name = property_name

    @property
    def has_value(self) -> bool:
        """Returns true if the device has a value property."""
        return self._property_name in self._properties

    @property
    def is_bool_value(self) -> bool:
        """Returns True if the device value property is a bool,
        either as string or real bool type.
        """
        if self.has_value:
            value = self._properties.get(self._property_name)
            if isinstance(value, bool):
                return True
            if isinstance(value, str):
                return value.lower() in ("true", "false")

        return False

    def str_value(self, default: str | None = None) -> str:
        """Returns the value converted to str or default if
        the object has no value.

        Raises:
        If no default is set, a TypeError is raised if no value exists.
        """
        if self.has_value:
            return str(self._properties.get(self._property_name, default))

        if default is None:
            raise TypeError("No value attribute available")
        return default

    def int_value(self, default: int | None = None) -> int:
        """Returns the value converted to int or default if
        conversion failes or the object has no value.

        Raises:
        If no default is set, a TypeError is raised for invalid values.
        """
        try:
            return int(float(self._properties.get(self._property_name)))
        except TypeError as ex:
            if default is None:
                raise ex
            return default

    def float_value(self, default: float | None = None) -> float:
        """Returns the value converted to float or default if
        conversion failes or the object has no value.

        Raises:
        If no default is set, a TypeError is raised for invalid values.
        """
        try:
            return float(self._properties.get(self._property_name))
        except TypeError as ex:
            if default is None:
                raise ex
            return default

    def bool_value(self, default: bool | None = None) -> bool:
        """Returns the value converted to bool or default if
        conversion failes or the object has no value.

        Raises:
        If no default is set, a TypeError is raised for invalid values.
        """
        try:
            value = self._properties.get(self._property_name)
            if isinstance(value, bool):
                return value
            if isinstance(value, str):
                if self.is_bool_value:
                    return value.lower() == "true"
                return self.float_value(0) != 0
            if isinstance(value, (int, float)):
                return value != 0
            raise TypeError(f"Value cannot be converted to bool {value}")
        except TypeError as ex:
            if default is None:
                raise ex
            return default

    def dict_value(self, default: dict | None = None) -> dict:
        """Returns the value converted to a dict or default if
        conversion failes or the object has no value.

        Raises:
        If no default is set, an error is raised for invalid values.
        """
        try:
            value = self._properties.get(self._property_name)
            if isinstance(value, dict):
                return value
            if isinstance(value, str):
                return json.loads(value)
            raise TypeError(f"Value cannot be converted to dict {value}")
        except TypeError as ex:
            if default is None:
                raise ex
            return default


class ColorModel:
    """Model to read out the color."""

    def __init__(self, properties: dict, property_name: str) -> None:
        """Constructor."""
        self._properties = properties
        self._property_name = property_name

    @property
    def has_color(self) -> bool:
        """Returns true if the device has a value property."""
        if self._property_name in self._properties:
            try:
                self.rgbw_color
                return True
            except (TypeError, ValueError):
                return False
        return False

    @property
    def rgbw_color(self) -> tuple[int, int, int, int]:
        """Returns the color as RGBW value.
        For RGB devices the white value is reported as 0.

        Raises:
        TypeError is raised for invalid values.
        """
        color = self._properties.get(self._property_name)
        if color is None:
            raise TypeError("Color is None.")
        rgbw = tuple(int(i) for i in color.split(","))
        if len(rgbw) != 4:
            raise TypeError(f"Color does not have 4 parts: {color}")
        return rgbw


class SceneEvent:
    """Model to read out the scene events."""

    def __init__(self, key_id: int, key_attributes: list[str]) -> None:
        """Constructor."""
        self._key_id = key_id
        self._key_attributes = key_attributes

    @property
    def key_id(self) -> int:
        """Returns the key id."""
        return self._key_id

    @property
    def key_event_types(self) -> list[str]:
        """Returns the possible key event types."""
        return self._key_attributes



"""Main class for accessing fibaro API."""

from requests import HTTPError

from .common.rest_client import RestClient
from .fibaro_device import DeviceModel
from .fibaro_info import InfoModel
from .fibaro_login import LoginModel
from .fibaro_room import RoomModel
from .fibaro_scene import SceneModel
from .fibaro_state_handler import FibaroStateHandler


class FibaroClient:
    """Fibaro client.

    This is the main entry point to access the fibaro API.

    Usage:
    Use set_authentication() to provide the credentials
    Use connect() to establish the connection and check if the credentials are valid
    Use any other method to access API data and actions
    """

    def __init__(self, url: str, ssl_verify: bool = False) -> None:
        """Init the fibaro client.

        The url needs to be in the format http(s)://<HOST>/api/.

        You can use ssl_verify to enable SSL certificate validation, but be
        aware that you need to register the fibaro certificates yourself
        to make it work. Also please be aware that the InsecureRequestWarning
        is not suppressed by default, so if you need to suppress warnings you
        need something like:

        import urllib3
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        """
        self._rest_client = RestClient(url, ssl_verify)
        self._frontend_url = url.removesuffix("/api/")
        self._api_version: int = None
        self._state_handler: FibaroStateHandler = None

    def set_authentication(self, username: str, password: str) -> None:
        """Set the credentials."""
        self._rest_client.set_auth(username, password)

    def connect(self) -> bool:
        """Returns the login status.

        Returns:
        True if authenticated, False if not authenticated

        Raises:
        HTTPError: If there is a connection problem. Most important is HTTPError
        with status 403 which raised if invalid credentials are provided.
        """
        login, _ = self._login()
        return login.is_logged_in

    def connect_with_credentials(self, username: str, password: str) -> InfoModel:
        """Connect with given credentials.
        Translate connect errors to easily differentiate auth and connect failures.

        Returns the hub info if successfully connected.
        Raises:
        FibaroAuthenticationFailed: If credentials are invalid
        FibaroConnectFailed: If connection is not possible
        """
        try:
            self.set_authentication(username, password)
            _, info = self._login()
            return info
        except HTTPError as http_ex:
            if http_ex.response.status_code == 403:
                raise FibaroAuthenticationFailed from http_ex
            raise FibaroConnectFailed from http_ex
        except Exception as ex:
            raise FibaroConnectFailed from ex

    def _login(self) -> tuple[LoginModel, InfoModel]:
        login = LoginModel(self._rest_client)
        info = self.read_info()

        # Read the API version as it is needed regularly
        self._api_version = info.api_version

        return (login, info)

    def read_info(self) -> InfoModel:
        """Read the info endpoint from home center."""
        return InfoModel(self._rest_client)

    def read_rooms(self) -> list[RoomModel]:
        """Read the rooms endpoint from home center."""
        return RoomModel.read_rooms(self._rest_client)

    def read_scenes(self) -> list[SceneModel]:
        """Read the scenes endpoint from home center."""
        return SceneModel.read_scenes(self._rest_client, self._api_version)

    def read_devices(self) -> list[DeviceModel]:
        """Read the devices endpoint from home center."""
        return DeviceModel.read_devices(self._rest_client, self._api_version)

    def register_update_handler(self, callback: callable) -> None:
        """Register a state handler."""
        if self._state_handler:
            raise Exception("There is already a state handler registered")
        self._state_handler = FibaroStateHandler(self._rest_client, callback)

    def unregister_update_handler(self) -> None:
        """Unregister the state handler."""
        if self._state_handler:
            self._state_handler.stop()
            self._state_handler = None

    def frontend_url(self) -> str:
        """Return the url to the web frontend of the fibaro hub."""
        return self._frontend_url


class FibaroConnectFailed(Exception):
    """Error to indicate we cannot connect to fibaro home center."""


class FibaroAuthenticationFailed(Exception):
    """Error to indicate that authentication failed on fibaro home center."""



"""Endpoint object to access the endpoint settings/info"""

import logging

from .common.const import (
    API_VERSION_MATCHER,
    MANUFACTURER_NAME_MATCHER,
    MODEL_NAME_MATCHER,
)
from .common.rest_client import RestClient

_LOGGER = logging.getLogger(__name__)


class InfoModel:
    """Fibaro info."""

    def __init__(self, rest_client: RestClient) -> None:
        """Load the data."""
        self.raw_data: dict = rest_client.get("settings/info")

    @property
    def current_version(self) -> str:
        """Returns the software version."""
        return self.raw_data.get("softVersion")

    @property
    def serial_number(self) -> str:
        """Returns the serial number of the home center."""
        return self.raw_data.get("serialNumber")

    @property
    def hc_name(self) -> str:
        """Returns the serial number of the home center."""
        return self.raw_data.get("hcName")

    @property
    def mac_address(self) -> str:
        """Returns the mac address of the home center."""
        return self.raw_data.get("mac")

    @property
    def api_version(self) -> int:
        """Returns the API version. As of writing version 4 and 5 was supported.

        When the version cannot be evaluated, it will fallback to version 5.
        """
        serial_number = self.serial_number
        for item in API_VERSION_MATCHER.items():
            if serial_number.startswith(item[0]):
                _LOGGER.debug(
                    "API version %s found by pattern %s", item[1], item[0])
                return item[1]
        return 5

    @property
    def platform(self) -> str:
        """Returns the model abbreviation of the home center.

        One of HC3, HC3L, YH, HC2 or HCL.
        """
        # This API is only available on newer models, therefore
        # we simulate an answer for older models.
        return self.raw_data.get("platform", self.serial_number[:3])

    @property
    def model_name(self) -> str:
        """Returns the name of the hub model.

        When the model cannot be evaluated, it will fallback to Hub.
        """
        serial_number = self.serial_number
        for item in MODEL_NAME_MATCHER.items():
            if serial_number.startswith(item[0]):
                return item[1]
        return "Hub"

    @property
    def manufacturer_name(self) -> str:
        """Returns the name of the hub model.

        When the model cannot be evaluated, it will fallback to Fibaro.
        """
        serial_number = self.serial_number
        for item in MANUFACTURER_NAME_MATCHER.items():
            if serial_number.startswith(item[0]):
                return item[1]
        return "Fibaro"



"""High level controller to access devices over the Fibaro API.

This adds a structural layer over the plain fibaro API which does a lot of
device detection logic and exposes properties in a more understandable way.
"""

from __future__ import annotations

import logging
from collections.abc import Callable

from .fibaro_state_multiplexer import FibaroStateMultiplexer
from .fibaro_client import FibaroClient
from .fibaro_device import DeviceModel
from .fibaro_state_resolver import FibaroEvent

_LOGGER = logging.getLogger(__name__)


class FibaroDeviceManager:
    """Controller to access fibaro API in a more structured way."""

    def __init__(
        self, fibaro_client: FibaroClient, include_devices_from_plugins: bool = False
    ) -> None:
        """Construct the fibaro device manager.
        - Load initial data
        - Open push channel"""
        self._fibaro_client = fibaro_client
        self._fibaro_state_multiplexer = FibaroStateMultiplexer(
            fibaro_client, include_devices_from_plugins
        )
        self._fibaro_state_multiplexer.start()

    def add_change_listener(
        self, fibaro_id: int, listener: Callable[[DeviceModel], None]
    ) -> Callable[[], None]:
        """Add a listener to get property changes.
        Provides the updated device data.

        Returns: Callback which can be used to unregister the listener"""
        return self._fibaro_state_multiplexer.add_change_listener(fibaro_id, listener)

    def add_event_listener(
        self, fibaro_id: int, listener: Callable[[FibaroEvent], None]
    ) -> Callable[[], None]:
        """Add scene event listener.

        Returns: Callback which can be used to unregister the listener"""
        return self._fibaro_state_multiplexer.add_event_listener(fibaro_id, listener)

    def get_devices(self) -> list[DeviceModel]:
        """Get current devices from Fibaro Home Center."""
        return self._fibaro_state_multiplexer.get_devices()

    def close(self) -> None:
        """Close push channel."""
        self._fibaro_state_multiplexer.stop()



"""Manual test class."""

import logging

from pyfibaro.fibaro_client import FibaroClient

from . import FIBARO_PASSWORD, FIBARO_URL, FIBARO_USERNAME


def main():
    """Main method for testing purposes"""

    logging.basicConfig(level=logging.DEBUG)

    print("Start test...")

    client = FibaroClient(FIBARO_URL)
    client.set_authentication(FIBARO_USERNAME, FIBARO_PASSWORD)
    client.connect()

    devices = client.read_devices()
    for device in devices:
        print(f"Device {device.fibaro_id}: {device.name}")
        if device.fibaro_id == 72:
            device.execute_action("turnOn")

        if device.fibaro_id == 347:
            device.execute_action("setValue", [99])


if __name__ == "__main__":
    main()



"""Manual test class."""

import logging

from pyfibaro.fibaro_client import FibaroClient

from . import FIBARO_PASSWORD, FIBARO_URL, FIBARO_USERNAME


def main():
    """Main method for testing purposes"""

    logging.basicConfig(level=logging.DEBUG)

    print("Start test...")

    client = FibaroClient(FIBARO_URL)
    logged_in = client.connect()

    client.set_authentication(FIBARO_USERNAME, FIBARO_PASSWORD)

    logged_in = client.connect()
    print(f"Logged in: {logged_in}")


if __name__ == "__main__":
    main()



"""Test FibaroInfo class."""

import requests_mock

from pyfibaro.fibaro_client import FibaroClient

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture


def test_fibaro_room() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        room_payload = load_fixture("room.json")
        mock.register_uri("GET", f"{TEST_BASE_URL}rooms", json=room_payload)
        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)

        rooms = client.read_rooms()
        assert rooms is not None

        assert len(rooms) == 2
        assert rooms[0].fibaro_id == 4
        assert rooms[0].name == "Wohnen"
        assert mock.call_count == 1



"""State handler for fibaro home center."""

import logging
import threading

from .common.const import REFRESH_STATE_TIMEOUT
from .common.rest_client import RestClient

_LOGGER = logging.getLogger(__name__)


class FibaroStateHandler(threading.Thread):
    """State handler which uses the refreshStates
    endpoint to pull state changes from home center.
    """

    def __init__(self, rest_client: RestClient, callback: callable) -> None:
        """Create the state handler and start the background thread."""

        super().__init__(name=f"Thread {__name__}")

        self._rest_client = rest_client
        self._callback = callback
        self._stop_flag = threading.Event()

        # stop unconditionally on exit
        self.daemon = True

        self.start()

    def run(self) -> None:
        """State Handler main loop which runs in this thread."""

        _LOGGER.info("Starting the state change handler")
        last = 0

        while not self._is_stopped_flag():
            sleep_time = 1
            attempt = 1
            success = False

            while not success and not self._is_stopped_flag():
                try:
                    state = self._rest_client.get(
                        f"refreshStates?last={last}", timeout=REFRESH_STATE_TIMEOUT
                    )
                    _LOGGER.debug(state)

                    last = state.get("last")
                    success = True
                except Exception as ex:
                    _LOGGER.warning("Connection Error (%s). Error: %s", attempt, ex)
                    attempt += 1

                    if attempt == 3:
                        sleep_time = 30
                        _LOGGER.info("Fallback to 30-second connection retry timer.")

                    self._stop_flag.wait(sleep_time)

                if success:
                    try:
                        self._callback(state)
                    except Exception as ex:
                        _LOGGER.warning("Error in state change callback: %s", ex)


        _LOGGER.info("State change handler stopped.")

    def _is_stopped_flag(self) -> bool:
        return self._stop_flag.is_set()

    def stop(self) -> None:
        """Stop the state handler."""
        _LOGGER.debug("Stopping the state change handler")

        # no effect on pending request
        self._rest_client.close()
        self._stop_flag.set()



"""Test rest client."""

import requests_mock

from pyfibaro.common.rest_client import RestClient

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture

info_payload = load_fixture("info.json")


def test_get_request() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=info_payload)
        client = RestClient(TEST_BASE_URL, False, TEST_USERNAME, TEST_PASSWORD)
        response = client.get("settings/info")

        assert mock.call_count == 1
        assert response == info_payload


def test_https_get_request() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)
        url = "https://192.169.1.40/api/"
        mock.register_uri("GET", f"{url}settings/info", json=info_payload)
        client = RestClient(url, True, TEST_USERNAME, TEST_PASSWORD)
        response = client.get("settings/info")

        assert mock.call_count == 1
        assert response == info_payload


def test_post_request() -> None:
    """Test post request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)
        mock.register_uri(
            "POST", f"{TEST_BASE_URL}settings/info", json=info_payload)
        client = RestClient(TEST_BASE_URL, False, TEST_USERNAME, TEST_PASSWORD)
        response = client.post("settings/info", info_payload)

        assert mock.call_count == 1
        assert response == info_payload



"""Test FibaroInfo class."""
from __future__ import annotations

from typing import Any

import pytest
import requests_mock

from pyfibaro.fibaro_client import FibaroClient
from pyfibaro.fibaro_device import ColorModel, DeviceModel, ValueModel

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture

device_payload = load_fixture("device.json")
login_payload = load_fixture("login_success.json")
info_payload = load_fixture("info.json")


def test_fibaro_device() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()
        assert devices is not None

        assert len(devices) == 3
        assert devices[0].fibaro_id == 1
        assert devices[0].name == "zwave"
        assert devices[0].parent_fibaro_id == 0
        assert devices[0].type == "com.fibaro.zwavePrimaryController"
        assert devices[0].base_type == ""
        assert devices[0].room_id == 0
        assert devices[0].has_unit is False
        assert devices[0].unit is None
        assert devices[0].has_armed is False
        assert devices[0].armed is False
        assert devices[0].has_interface("invalid_interfce") is False
        assert devices[2].has_interface("energy") is True
        assert devices[0].has_battery_level is False
        assert devices[0].battery_level == 0
        assert devices[0].has_endpoint_id is True
        assert devices[0].endpoint_id == 0
        assert devices[0].enabled is True
        assert devices[0].visible is False
        assert devices[0].is_plugin is False
        assert devices[0].value.int_value() == 0
        assert devices[0].value_2.has_value is False
        assert devices[0].color.has_color is False
        assert devices[0].last_color_set.has_color is False
        assert devices[0].state.has_value is False
        assert devices[0].has_brightness is False
        assert devices[0].brightness == 0
        assert devices[0].current_program == 0
        assert devices[0].current_program_id == 0
        assert devices[0].has_mode is False
        assert devices[0].mode == 0
        assert devices[0].has_supported_modes is False
        assert devices[0].supported_modes == []
        assert devices[0].has_operating_mode is False
        assert devices[0].operating_mode == 0
        assert devices[0].has_supported_operating_modes is False
        assert devices[0].supported_operating_modes == []
        assert devices[0].has_thermostat_mode is False
        assert devices[0].thermostat_mode is None
        assert devices[0].has_thermostat_operating_state is False
        assert devices[0].thermostat_operating_state is None
        assert devices[0].has_supported_thermostat_modes is False
        assert devices[0].supported_thermostat_modes == []
        assert devices[0].has_heating_thermostat_setpoint is False
        assert devices[0].heating_thermostat_setpoint == 0
        assert devices[0].has_heating_thermostat_setpoint_future is False
        assert devices[0].heating_thermostat_setpoint_future == 0
        assert devices[0].target_level == 0
        assert devices[0].has_dead is True
        assert devices[0].dead is False
        assert devices[0].has_dead_reason is False
        assert devices[0].dead_reason is None

        assert isinstance(devices[0].actions, dict)
        assert isinstance(devices[0].properties, dict)
        assert mock.call_count == 3


@pytest.mark.parametrize("test_value", ["true", "True", True, 1, 0.1, "1", "0.1"])
def test_fibaro_value_true(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.has_value is True
    assert value.bool_value() is True
    assert value.bool_value(False) is True


@pytest.mark.parametrize("test_value", ["false", "False", False, 0, 0.0, "0", "0.0"])
def test_fibaro_value_false(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.has_value is True
    assert value.bool_value() is False
    assert value.bool_value(True) is False


@pytest.mark.parametrize("test_value", ["1", "1.1", 1, 1.1])
def test_fibaro_value_int(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.has_value is True
    assert value.int_value() == 1


@pytest.mark.parametrize("test_value", ["1", 1])
def test_fibaro_value_float(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.has_value is True
    assert value.float_value() == 1


@pytest.mark.parametrize("test_value", ["1", 1])
def test_fibaro_value_str(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.has_value is True
    assert value.str_value() == "1"


@pytest.mark.parametrize("test_value", [{"x": 1}, '{"x":1}'])
def test_fibaro_value_dict(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.has_value is True
    assert value.dict_value() == {"x": 1}


@pytest.mark.parametrize("test_value", ["1.1", 1.1])
def test_fibaro_value_float_2(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.has_value is True
    assert value.float_value() == 1.1


def test_fibaro_value_default() -> None:
    """Test value model"""
    value = ValueModel({}, "value")
    assert value.has_value is False
    assert value.int_value(1) == 1
    assert value.float_value(1.1) == 1.1
    assert value.bool_value(True) is True
    assert value.str_value("test") == "test"
    assert value.dict_value({"x": 1}) == {"x": 1}


@pytest.mark.parametrize("test_value", ["true", "True", True, "false", "False", False])
def test_fibaro_value_is_bool(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.is_bool_value is True


@pytest.mark.parametrize("test_value", [0, 1, "any_value"])
def test_fibaro_value_is_bool_negative(test_value: Any) -> None:
    """Test value model"""
    value = ValueModel({"value": test_value}, "value")
    assert value.is_bool_value is False


def test_fibaro_no_value() -> None:
    """Test value model"""
    value = ValueModel({}, "value")
    assert value.has_value is False
    with pytest.raises(TypeError):
        value.int_value()
    with pytest.raises(TypeError):
        value.float_value()
    with pytest.raises(TypeError):
        value.bool_value()
    with pytest.raises(TypeError):
        value.str_value()
    with pytest.raises(TypeError):
        value.dict_value()


def test_fibaro_color() -> None:
    """Test value model"""
    value = ColorModel({"color": "0,0,0,0"}, "color")
    assert value.has_color is True
    assert value.rgbw_color == (0, 0, 0, 0)

    value = ColorModel({"color": "234,244,255,0"}, "color")
    assert value.has_color is True
    assert value.rgbw_color == (234, 244, 255, 0)

    value = ColorModel({"color": "234,255,0"}, "color")
    assert value.has_color is False
    with pytest.raises(TypeError):
        value.rgbw_color

    value = ColorModel({"color": "234,255,0"}, "color_xxxx")
    assert value.has_color is False
    with pytest.raises(TypeError):
        value.rgbw_color


@pytest.mark.parametrize("test_value", ["1,2,3,4", ["1", "2", "3", "4"], [1, 2, 3, 4]])
def test_fibaro_supported_modes(test_value: Any) -> None:
    """Test modes"""
    device = DeviceModel({"properties": {"supportedModes": test_value}}, None, 4)
    assert device.supported_modes == [1, 2, 3, 4]


@pytest.mark.parametrize("test_value", ["1,2,3,4", ["1", "2", "3", "4"], [1, 2, 3, 4]])
def test_fibaro_supported_operating_modes(test_value: Any) -> None:
    """Test operating modes"""
    device = DeviceModel(
        {"properties": {"supportedOperatingModes": test_value}}, None, 4
    )
    assert device.supported_operating_modes == [1, 2, 3, 4]


def test_fibaro_supported_thermostat_modes() -> None:
    """Test thermostat modes"""
    device = DeviceModel(
        {"properties": {"supportedThermostatModes": ["heat", "auto"]}}, None, 4
    )
    assert device.supported_thermostat_modes == ["heat", "auto"]


def test_fibaro_device_turn_on() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}devices/13/action/turnOn")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()

        assert mock.call_count == 3
        devices[2].execute_action("turnOn")
        assert mock.call_count == 4


def test_fibaro_device_invalid_action() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}devices/13/action/XXXX")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()

        assert mock.call_count == 3
        devices[2].execute_action("XXXX")
        assert mock.call_count == 4


def test_fibaro_device_action_with_param() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}devices/13/action/abortUpdate")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()

        assert mock.call_count == 3
        devices[2].execute_action("abortUpdate", [True])
        assert mock.call_count == 4



"""Test FibaroInfo class."""

import requests_mock

from pyfibaro.fibaro_client import FibaroClient

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture

device_payload = load_fixture("device-armed.json")
login_payload = load_fixture("login_success.json")
info_payload = load_fixture("info.json")


def test_fibaro_device_armed() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()
        assert devices is not None

        assert len(devices) == 3

        assert devices[1].has_armed is True
        # Invalid value returns False
        assert devices[1].armed is False

        assert devices[2].has_armed is True
        assert devices[2].armed is True



"""Test FibaroStateHandler class."""

import time

import pytest
import requests_mock

from pyfibaro.fibaro_client import FibaroClient

from .test_utils import (TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME,
                         load_fixture)

refresh_payload = load_fixture("refresh.json")
login_payload = load_fixture("login_success.json")
info_payload = load_fixture("info.json")


class TestRefresh:
    """Use a test class to get async state."""

    callback_result = None

    def callback_function(self, event) -> None:
        """The callback handler for this test."""
        self.callback_result = event

    def callback_exception(self, event) -> None:
        """The callback handler for this test."""
        self.callback_result = event
        raise TypeError()

    def test_fibaro_refresh(self) -> None:
        """Test get request"""
        with requests_mock.Mocker() as mock:
            assert isinstance(mock, requests_mock.Mocker)

            mock.register_uri(
                "GET", f"{TEST_BASE_URL}refreshStates", json=refresh_payload
            )
            mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
            mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

            client = FibaroClient(TEST_BASE_URL)
            client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
            client.connect()

            client.register_update_handler(self.callback_function)
            time.sleep(0.1)
            client.unregister_update_handler()

            assert self.callback_result is not None
            assert mock.call_count > 2

    def test_fibaro_refresh_fail(self) -> None:
        """Test get request"""
        with requests_mock.Mocker() as mock:
            assert isinstance(mock, requests_mock.Mocker)

            mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
            mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

            client = FibaroClient(TEST_BASE_URL)
            client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
            client.connect()

            client.register_update_handler(self.callback_function)
            time.sleep(1.1)
            client.unregister_update_handler()

            assert self.callback_result is None
            assert mock.call_count > 2

    def test_fibaro_refresh_double_register(self) -> None:
        """Test get request"""
        with requests_mock.Mocker() as mock:
            assert isinstance(mock, requests_mock.Mocker)

            mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
            mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

            client = FibaroClient(TEST_BASE_URL)
            client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
            client.connect()
            client.register_update_handler(self.callback_function)

            with pytest.raises(Exception):
                client.register_update_handler(self.callback_function)

            client.unregister_update_handler()

    def test_fibaro_refresh_callback_exception(self) -> None:
        """Test get request"""
        with requests_mock.Mocker() as mock:
            assert isinstance(mock, requests_mock.Mocker)

            mock.register_uri(
                "GET", f"{TEST_BASE_URL}refreshStates", json=refresh_payload
            )
            mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
            mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

            client = FibaroClient(TEST_BASE_URL)
            client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
            client.connect()

            client.register_update_handler(self.callback_exception)
            time.sleep(0.1)
            client.unregister_update_handler()

            assert self.callback_result is not None
            assert mock.call_count > 2



"""Test FibaroStateMultiplexer class."""

from unittest.mock import Mock
from pyfibaro.fibaro_device import DeviceModel
from pyfibaro.fibaro_state_multiplexer import FibaroStateMultiplexer

from .test_utils import load_fixture

refresh_payload = load_fixture("refresh.json")
device_payload = load_fixture("device.json")


def test_fibaro_state_multiplexer_start() -> None:
    """Test state multiplexer start."""
    devices = [
        DeviceModel(device_payload[2], Mock(), 4),
        DeviceModel(device_payload[3], Mock(), 4),
    ]

    fibaro_client = Mock()
    fibaro_client.read_devices.return_value = devices
    multiplexer = FibaroStateMultiplexer(fibaro_client)

    multiplexer.start()

    fibaro_client.register_update_handler.assert_called_once()


def test_fibaro_state_multiplexer_stop() -> None:
    """Test state multiplexer stop."""
    fibaro_client = Mock()
    multiplexer = FibaroStateMultiplexer(fibaro_client)

    multiplexer.stop()

    fibaro_client.unregister_update_handler.assert_called_once()


def test_fibaro_state_multiplexer_state_change_listener() -> None:
    """Test state multiplexer add state change listener."""
    devices = [
        DeviceModel(device_payload[2], Mock(), 4),
        DeviceModel(device_payload[3], Mock(), 4),
    ]

    fibaro_client = Mock()
    fibaro_client.read_devices.return_value = devices

    multiplexer = FibaroStateMultiplexer(fibaro_client)
    multiplexer.start()

    result_mock = Mock()
    multiplexer.add_change_listener(13, result_mock.call_method)
    assert len(multiplexer._change_listeners.get(13)) == 1

    multiplexer._on_change(refresh_payload)

    result_mock.call_method.assert_called_once()


def test_fibaro_state_multiplexer_get_devices() -> None:
    """Test state multiplexer add state change listener."""
    devices = [
        DeviceModel(device_payload[2], Mock(), 4),
        DeviceModel(device_payload[3], Mock(), 4),
    ]

    fibaro_client = Mock()
    fibaro_client.read_devices.return_value = devices

    multiplexer = FibaroStateMultiplexer(fibaro_client)
    multiplexer.start()
    devices = multiplexer.get_devices()
    assert len(devices) == 2


def test_fibaro_state_multiplexer_add_remove_state_change_listener() -> None:
    """Test state multiplexer add state change listener."""
    devices = [
        DeviceModel(device_payload[2], Mock(), 4),
        DeviceModel(device_payload[3], Mock(), 4),
    ]

    fibaro_client = Mock()
    fibaro_client.read_devices.return_value = devices

    multiplexer = FibaroStateMultiplexer(fibaro_client)

    result_mock = Mock()
    remove = multiplexer.add_change_listener(13, result_mock.call_method)
    assert len(multiplexer._change_listeners.get(13)) == 1

    remove()

    assert len(multiplexer._change_listeners.get(13)) == 0


def test_fibaro_state_multiplexer_event_listener() -> None:
    """Test state multiplexer add event listener."""

    fibaro_client = Mock()
    multiplexer = FibaroStateMultiplexer(fibaro_client)

    result_mock = Mock()
    multiplexer.add_event_listener(28, result_mock.call_method)

    multiplexer._on_change(refresh_payload)

    result_mock.call_method.assert_called_once()



"""Manual test class."""

import logging

from pyfibaro.fibaro_client import FibaroClient

from . import FIBARO_URL


def main():
    """Main method for testing purposes"""

    logging.basicConfig(level=logging.DEBUG)

    print("Start test...")

    client = FibaroClient(FIBARO_URL)
    info = client.read_info()

    print(f"Serial no: {info.serial_number}")
    print(f"Version: {info.current_version}")
    print(f"HC Name: {info.hc_name}")
    print(f"API version: {info.api_version}")


if __name__ == "__main__":
    main()



"""Manual test class."""

import logging
import time

from pyfibaro.fibaro_client import FibaroClient

from . import FIBARO_PASSWORD, FIBARO_URL, FIBARO_USERNAME


def state_callback(event) -> None:
    """Report state updates to the console."""
    print(f"State updates: {event}")


def main():
    """Main method for testing purposes"""

    logging.basicConfig(level=logging.DEBUG)

    print("Start test...")

    client = FibaroClient(FIBARO_URL)
    client.set_authentication(FIBARO_USERNAME, FIBARO_PASSWORD)
    client.connect()

    client.register_update_handler(state_callback)

    time.sleep(60)


if __name__ == "__main__":
    main()



"""Test FibaroStateResolver class."""


from pyfibaro.fibaro_state_resolver import FibaroStateResolver

from .test_utils import load_fixture

refresh_payload = load_fixture("refresh.json")
refresh_extended_payload = load_fixture("refresh_extended.json")


def test_fibaro_state_resolver_event() -> None:
    """Test event resolver"""
    resolver = FibaroStateResolver(refresh_payload)
    events = resolver.get_events()
    assert events is not None
    assert len(events) == 2
    assert events[0].event_type == "DevicePropertyUpdatedEvent"
    assert events[0].fibaro_id == 28
    assert events[0].key_event_type == ""
    assert events[0].key_id == 0

    assert events[1].event_type == "PowerMetricsChangedEvent"
    assert events[1].fibaro_id is None
    assert events[1].key_event_type == ""
    assert events[1].key_id == 0


def test_fibaro_state_resolver_state_update() -> None:
    """Test event resolver"""
    resolver = FibaroStateResolver(refresh_payload)
    changes = resolver.get_state_updates()
    assert changes is not None
    assert len(changes) == 2
    assert changes[0].fibaro_id == 28
    assert changes[0].property_changes == {"value": "232.88"}


def test_fibaro_state_resolver_state_update_extended() -> None:
    """Test event resolver"""
    resolver = FibaroStateResolver(refresh_extended_payload)
    changes = resolver.get_state_updates()
    assert changes is not None
    assert len(changes) == 1
    assert changes[0].fibaro_id == 28
    assert changes[0].property_changes == {"value": "232.88"}



"""Test FibaroInfo class."""

import requests_mock

from pyfibaro.fibaro_client import FibaroClient

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture

device_payload = load_fixture("device-hc3.json")
login_payload = load_fixture("login_success.json")
info_payload = load_fixture("info.json")


def test_fibaro_device() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()
        assert devices is not None

        assert len(devices) == 4
        assert devices[0].fibaro_id == 1
        assert devices[0].name == "zwave"
        assert devices[0].parent_fibaro_id == 0
        assert devices[0].type == "com.fibaro.zwavePrimaryController"
        assert devices[0].base_type == ""
        assert devices[0].room_id == 219
        assert isinstance(devices[0].actions, dict)
        assert isinstance(devices[0].properties, dict)
        assert mock.call_count == 3
        assert devices[0].has_dead is True
        assert devices[0].dead is True
        assert devices[0].has_dead_reason is True
        assert devices[0].dead_reason == "Connection problem"


def test_fibaro_device_turn_on() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}devices/7/action/turnOn")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()

        assert mock.call_count == 3
        devices[2].execute_action("turnOn")
        assert mock.call_count == 4


def test_fibaro_device_invalid_action() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}devices/7/action/XXXX")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()

        assert mock.call_count == 3
        devices[2].execute_action("XXXX")
        assert mock.call_count == 4


def test_fibaro_device_action_with_param() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}devices", json=device_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}devices/7/action/abortUpdate")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        devices = client.read_devices()

        assert mock.call_count == 3
        devices[2].execute_action("abortUpdate", [True])
        assert mock.call_count == 4



"""Manual test class."""

import logging

from pyfibaro.fibaro_client import FibaroClient

from . import FIBARO_PASSWORD, FIBARO_URL, FIBARO_USERNAME


def main():
    """Main method for testing purposes"""

    logging.basicConfig(level=logging.DEBUG)

    print("Start test...")

    client = FibaroClient(FIBARO_URL)
    client.set_authentication(FIBARO_USERNAME, FIBARO_PASSWORD)
    client.connect()

    scenes = client.read_scenes()
    for scene in scenes:
        print(f"Scene {scene.fibaro_id}: {scene.name}")

        if scene.fibaro_id == 19:
            scene.start()
            scene.stop()


if __name__ == "__main__":
    main()



"""State object resolver for fibaro home center."""
from __future__ import annotations

import logging

from typing import Any

_LOGGER = logging.getLogger(__name__)


class FibaroEvent:
    """A fibaro event returned by state handler."""

    def __init__(self, data: dict) -> None:
        """Constructor to init the object with the raw data."""
        self.raw_data = data

    @property
    def event_type(self) -> str:
        """Returns the event type as string.

        Most known event is "CentralSceneEvent" which is used for button press events.
        """
        return self.raw_data.get("type", "")

    @property
    def fibaro_id(self) -> int | None:
        """The device id which throws the event. Not all events are related to a device."""
        data = self.event_data
        # id is used by HC3, deviceId by HC2
        fibaro_id = data.get("id", data.get("deviceId"))
        if fibaro_id is None:
            return None
        return int(fibaro_id)

    @property
    def event_data(self) -> dict:
        """Returns the event data in raw format."""
        return self.raw_data.get("data", {})

    @property
    def key_id(self) -> int:
        """Returns the key id."""
        return int(self.event_data.get("keyId", 0))

    @property
    def key_event_type(self) -> str:
        """Returns the key event attribute.

        For example Pressed, Released, HeldDown, ...
        """
        return self.event_data.get("keyAttribute", "")


class FibaroStateChange:
    """A fibaro state change returned by state handler."""

    def __init__(self, data: dict) -> None:
        """Constructor to init the object with the raw data."""
        self.raw_data = data

    @property
    def fibaro_id(self) -> int:
        """The device id which throws the event."""
        return int(self.raw_data.get("id"))

    @property
    def property_changes(self) -> dict[str, Any]:
        """The changes in the device properties."""
        result: dict[str, Any] = {}

        for property_name, value in self.raw_data.items():
            # Ignore some attributes which are not relevant or returned separately
            if property_name in ("log", "logTemp", "id"):
                continue
            result[property_name] = value

        return result


class FibaroStateResolver:
    """State resolver allows typed access to a state object retunred by fibaro home center."""

    def __init__(self, data: dict) -> None:
        """Init the object with the raw event object."""
        self.raw_data = data

    def get_events(self) -> list[FibaroEvent]:
        """Extract events from the state handle object."""
        return [FibaroEvent(data) for data in self.raw_data.get("events", [])]

    def get_state_updates(self) -> list[FibaroStateChange]:
        """Extract state changes from the state handle object."""
        return [FibaroStateChange(data) for data in self.raw_data.get("changes", [])]



"""Test fibaro connection."""

import pytest
import requests_mock

from pyfibaro.fibaro_client import FibaroClient
from pyfibaro.fibaro_client import (
    FibaroAuthenticationFailed,
    FibaroConnectFailed
)

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture

login_payload = load_fixture("login_success.json")
info_payload = load_fixture("info.json")


def test_connect_failed_with_exception() -> None:
    """Test any exception."""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)

        with pytest.raises(FibaroConnectFailed):
            client.connect_with_credentials(TEST_USERNAME, TEST_PASSWORD)


def test_connect_failed_with_http_error() -> None:
    """Test http error."""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}loginStatus", status_code=500)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)

        with pytest.raises(FibaroConnectFailed):
            client.connect_with_credentials(TEST_USERNAME, TEST_PASSWORD)


def test_invalid_authentication() -> None:
    """Test invalid password."""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}loginStatus", status_code=403)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)

        with pytest.raises(FibaroAuthenticationFailed):
            client.connect_with_credentials(TEST_USERNAME, TEST_PASSWORD)


def test_connect_succeed() -> None:
    """Test connect succeeds."""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        info = client.connect_with_credentials(TEST_USERNAME, TEST_PASSWORD)

        assert mock.call_count == 2
        assert info.raw_data == info_payload


def test_frontend_url() -> None:
    """Test frontend url getter."""
    client = FibaroClient(TEST_BASE_URL)
    assert client.frontend_url() == TEST_BASE_URL.removesuffix("/api/")



"""Examples for using the API.

To try out set here your data and start the example you are interested in with
python -m examples.test_info
"""

FIBARO_URL = "http://192.168.1.166/api/"
FIBARO_USERNAME = "your_fibaro_username"
FIBARO_PASSWORD = "your_fibaro_password"



"""Endpoint object to access the endpoint settings/info"""
from __future__ import annotations

import logging

from .common.rest_client import RestClient

_LOGGER = logging.getLogger(__name__)


class SceneModel:
    """Model of a scene."""

    def __init__(self, data: dict, rest_client: RestClient, api_version: int) -> None:
        """One scene."""
        self.raw_data = data
        self._rest_client = rest_client
        self._api_version = api_version

    @property
    def fibaro_id(self) -> int:
        """Scene id"""
        return self.raw_data.get("id")

    @property
    def name(self) -> str:
        """Scene name"""
        return self.raw_data.get("name")

    @property
    def room_id(self) -> int:
        """Room id of the scene or 0 if no room is assigned."""
        if self._api_version == 4:
            return int(self.raw_data.get("roomID", 0))
        else:
            return int(self.raw_data.get("roomId", 0))

    @property
    def visible(self) -> bool:
        """Returns the visible state of the scene."""
        if self._api_version == 4:
            return self.raw_data.get("visible", True)
        else:
            return not self.raw_data.get("hidden", False)

    def start(self, user_pin: str | None = None) -> None:
        """Start a scene."""
        if self._api_version == 4:
            if user_pin:
                raise NotImplementedError("Not supported on old fibaro hubs")
            self._send_action_v4("start")
        else:
            self._send_action_v5("execute", user_pin)

    def stop(self, user_pin: str | None = None) -> None:
        """Stop a scene."""
        if self._api_version == 4:
            if user_pin:
                raise NotImplementedError("Not supported on old fibaro hubs")
            self._send_action_v4("stop")
        else:
            self._send_action_v5("kill", user_pin)

    def _send_action_v4(self, action: str) -> None:
        url = f"scenes/{self.fibaro_id}/action/{action}"
        self._rest_client.post(url)

    def _send_action_v5(self, action: str, user_pin: str | None) -> None:
        url = f"scenes/{self.fibaro_id}/{action}"
        if user_pin:
            self._rest_client.post(url, {}, http_headers={"Fibaro-User-PIN": user_pin})
        else:
            self._rest_client.post(url, {})

    @staticmethod
    def read_scenes(rest_client: RestClient, api_version: int) -> list[SceneModel]:
        """Returns a list of scenes."""
        raw_data: list = rest_client.get("scenes")
        scenes: list[dict] = []
        for scene in raw_data:
            if "id" not in scene or "name" not in scene:
                _LOGGER.debug("Ignore scene because it does not contain id or name")
            else:
                scenes.append(scene)

        return [SceneModel(data, rest_client, api_version) for data in scenes]



"""Test SceneModel class."""

import pytest
import requests_mock

from pyfibaro.fibaro_client import FibaroClient

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture

scene_payload = load_fixture("scene.json")
login_payload = load_fixture("login_success.json")
info_payload = load_fixture("info.json")


def test_fibaro_scene() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}scenes", json=scene_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        scenes = client.read_scenes()
        assert scenes is not None

        assert len(scenes) == 2
        assert scenes[0].fibaro_id == 2
        assert scenes[0].name == "Go to bed"
        assert scenes[0].room_id == 5
        assert scenes[0].visible is True
        assert mock.call_count == 3


def test_fibaro_scene_start() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}scenes", json=scene_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}scenes/2/action/start")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        scenes = client.read_scenes()

        assert mock.call_count == 3
        scenes[0].start()
        assert mock.call_count == 4


def test_fibaro_scene_start_with_pin_raise_exception() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}scenes", json=scene_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}scenes/2/action/start")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        scenes = client.read_scenes()

        assert mock.call_count == 3

        with pytest.raises(NotImplementedError):
            scenes[0].start("1234")
        assert mock.call_count == 3


def test_fibaro_scene_stop() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}scenes", json=scene_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}scenes/2/action/stop")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        scenes = client.read_scenes()

        assert mock.call_count == 3
        scenes[0].stop()
        assert mock.call_count == 4


def test_fibaro_scene_stop_with_pin_raises_exception() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri("GET", f"{TEST_BASE_URL}scenes", json=scene_payload)
        mock.register_uri("POST", f"{TEST_BASE_URL}scenes/2/action/stop")
        mock.register_uri("GET", f"{TEST_BASE_URL}loginStatus", json=login_payload)
        mock.register_uri("GET", f"{TEST_BASE_URL}settings/info", json=info_payload)

        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)
        client.connect()

        scenes = client.read_scenes()

        assert mock.call_count == 3
        with pytest.raises(NotImplementedError):
            scenes[0].stop("1234")
        assert mock.call_count == 3



"""Test FibaroInfo class."""

import pytest
import requests_mock

from pyfibaro.common.rest_client import RestClient
from pyfibaro.fibaro_client import FibaroClient
from pyfibaro.fibaro_info import InfoModel

from .test_utils import TEST_BASE_URL, TEST_PASSWORD, TEST_USERNAME, load_fixture

info_payload = load_fixture("info.json")


def test_extract_info() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=info_payload)
        client = RestClient(TEST_BASE_URL, TEST_USERNAME, TEST_PASSWORD)

        fibaro_info = InfoModel(client)

        assert mock.call_count == 1
        assert fibaro_info.raw_data == info_payload
        assert fibaro_info.current_version == "4.630"
        assert fibaro_info.hc_name == "My Home"
        assert fibaro_info.serial_number == "HC2-011111"
        assert fibaro_info.platform == "HC2"
        assert fibaro_info.mac_address == "00:22:4d:b7:13:24"


@pytest.mark.parametrize(
    "serial_number,api_version", [
        ("HC3-111111", 5), ("HC2-111111", 4), ("INVALID", 5)]
)
def test_api_version(serial_number: str, api_version: int) -> None:
    """Test API version"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        test_payload_copy = info_payload.copy()
        test_payload_copy["serialNumber"] = serial_number
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=test_payload_copy
        )
        client = RestClient(TEST_BASE_URL, TEST_USERNAME, TEST_PASSWORD)

        fibaro_info = InfoModel(client)

        assert mock.call_count == 1
        assert fibaro_info.api_version == api_version


@pytest.mark.parametrize(
    "serial_number,name",
    [
        ("HC3-111111", "Home Center 3"),
        ("HC2-111111", "Home Center 2"),
        ("INVALID", "Hub"),
    ],
)
def test_model_name_resolution(serial_number: str, name: int) -> None:
    """Test name resolution"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        test_payload_copy = info_payload.copy()
        test_payload_copy["serialNumber"] = serial_number
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=test_payload_copy
        )
        client = RestClient(TEST_BASE_URL, TEST_USERNAME, TEST_PASSWORD)

        fibaro_info = InfoModel(client)

        assert mock.call_count == 1
        assert fibaro_info.model_name == name


@pytest.mark.parametrize(
    "serial_number,name",
    [
        ("HC3-111111", "Fibaro"),
        ("HC2-111111", "Fibaro"),
        ("ZB", "ZOOZ"),
        ("INVALID", "Fibaro"),
    ],
)
def test_manufacturer_name_resolution(serial_number: str, name: int) -> None:
    """Test name resolution"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        test_payload_copy = info_payload.copy()
        test_payload_copy["serialNumber"] = serial_number
        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=test_payload_copy
        )
        client = RestClient(TEST_BASE_URL, TEST_USERNAME, TEST_PASSWORD)

        fibaro_info = InfoModel(client)

        assert mock.call_count == 1
        assert fibaro_info.manufacturer_name == name


def test_fibaro_info() -> None:
    """Test get request"""
    with requests_mock.Mocker() as mock:
        assert isinstance(mock, requests_mock.Mocker)

        mock.register_uri(
            "GET", f"{TEST_BASE_URL}settings/info", json=info_payload)
        client = FibaroClient(TEST_BASE_URL)
        client.set_authentication(TEST_USERNAME, TEST_PASSWORD)

        assert client.read_info() is not None
        assert mock.call_count == 1



"""Fibaro state multiplexer receives all information from the push channel and
provides methods to register listeners for specific devices.
"""

import logging
from typing import Any
from collections.abc import Callable

from .fibaro_client import FibaroClient
from .fibaro_device import DeviceModel
from .fibaro_data_helper import read_devices
from .fibaro_state_resolver import FibaroEvent, FibaroStateChange, FibaroStateResolver


_LOGGER = logging.getLogger(__name__)


class FibaroStateMultiplexer:
    """State and event multiplexer."""

    def __init__(
        self, fibaro_client: FibaroClient, include_devices_from_plugins: bool = False
    ) -> None:
        """Initialize the fibaro state multiplexer."""
        self._fibaro_client = fibaro_client
        self._include_devices_from_plugins = include_devices_from_plugins
        self._devices: dict[int, DeviceModel] = {}

        self._change_listeners: dict[int,
                                     list[Callable[[DeviceModel], None]]] = {}
        self._event_listeners: dict[int,
                                    list[Callable[[FibaroEvent], None]]] = {}

    def start(self) -> None:
        """Connect push channel and load initial device state.
        This starts change and event dispatching."""
        self._devices = {
            device.fibaro_id: device
            for device in read_devices(
                self._fibaro_client, self._include_devices_from_plugins
            )
        }
        self._fibaro_client.register_update_handler(self._on_change)

    def stop(self) -> None:
        """Disconnect push channel so that no change and events are dispatched anymore."""
        self._fibaro_client.unregister_update_handler()
        self._devices = {}

    def add_change_listener(
        self, fibaro_id: int, listener: Callable[[DeviceModel], None]
    ) -> Callable[[], None]:
        """Add a listener to get property changes."""
        change_listeners = self._change_listeners.setdefault(fibaro_id, [])
        change_listeners.append(listener)

        return lambda: change_listeners.remove(listener)

    def add_event_listener(
        self, fibaro_id: int, listener: Callable[[FibaroEvent], None]
    ) -> Callable[[], None]:
        """Add event listener."""
        event_listeners = self._event_listeners.setdefault(fibaro_id, [])
        event_listeners.append(listener)

        return lambda: event_listeners.remove(listener)

    def get_devices(self) -> list[DeviceModel]:
        """Return the current device state."""
        return list(self._devices.values())

    def _on_change(self, state: Any) -> None:
        # update internal device model and notify registered listeners
        resolver = FibaroStateResolver(state)

        for state_change in resolver.get_state_updates():
            fibaro_id = state_change.fibaro_id
            device = self._devices.get(fibaro_id)
            if device:
                self._update_device_data(device, state_change)
                for listener in self._change_listeners.get(fibaro_id, []):
                    listener(device)

        for event in resolver.get_events():
            # event does not always have a fibaro id, therefore it is
            # essential that we first check for it
            fibaro_id = event.fibaro_id
            if fibaro_id:
                for listener in self._event_listeners.get(fibaro_id, []):
                    listener(event)

    def _update_device_data(
        self, device: DeviceModel, state_change: FibaroStateChange
    ) -> None:
        # update the internal data object to keep it always current
        for key, value in state_change.property_changes.items():
            device.properties[key] = value
            _LOGGER.debug(
                "New state %s[%s].%s = %s", device.name, device.fibaro_id, key, str(
                    value)
            )



"""Constants for the fibaro client."""

# Timeout in seconds for http requests
DEFAULT_TIMEOUT = 10

# HC 3 needs a big timeout for refresh states request,
# it waits up to 30 seconds before a response is sent
REFRESH_STATE_TIMEOUT = 35

# Constant http headers sent with each request
HTTP_HEADERS = {
    "Content-Type": "application/json; charset=utf-8",
    "User-Agent": "pyfibaro",
}

# Match serial number prefix to a interface version
API_VERSION_MATCHER = {"HC2": 4, "HCL": 4,
                       "HC3": 5, "HC3L": 5, "YH": 5, "ZB": 5}

# Match serial number prefix to the model name
MODEL_NAME_MATCHER = {"HC2": "Home Center 2", "HCL": "Home Center Lite",
                      "HC3": "Home Center 3", "HC3L": "Home Center Lite",
                      "YH": "Yubii Home", "ZB": "Z-Box Hub"}

# Match serial number prefix to the manufacturer
MANUFACTURER_NAME_MATCHER = {"HC2": "Fibaro", "HCL": "Fibaro",
                             "HC3": "Fibaro", "HC3L": "Fibaro",
                             "YH": "Yubii", "ZB": "ZOOZ"}

# Devices which are ignored
# iOS_device includes iOS and Android devices
IGNORE_DEVICE = ["HC_user", "VOIP_user", "iOS_device"]
